(* Extra definitions and lemmas about boolean relation on finite types *)

Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq path fintype.
Require Import choice finset fingraph.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Section ExtraDefs.

Variable F : finType.

Definition rel_seq (r1 r2 : rel F) := [rel x y| existsb z, r1 x z && r2 z y].

(* This definition is not the most efficient one, do we need localisation at*)
(* all? see the equivalence lemma acyclic_restr below *)

Definition acyclic_on (s : {set F}) (e : rel F) :=
  forall (c : seq F), size c > 0 -> c \subset s -> ~ cycle e c.

Definition acyclic := acyclic_on setT.

Definition total_irr (e : rel F) := forall x y, [|| x == y, e x y | e y x].

Definition asymmetric (e : rel F) := forall a b, ~~(e a b && e b a).

Definition domUrange (r : rel F) :=
  [set x | existsb y, r x y] :|: [set y | existsb x, r x y].

Definition connect_irr (r : rel F) := [rel x y | (x != y) && (connect r x y)].

End ExtraDefs.


Section ExtraRelLemmas.

Variables (F : finType)(r : rel F).
Implicit Types r : rel F.

Lemma dom_domUrangeS : [set y | existsb x, r x y] \subset domUrange r.
Proof. exact: subsetUr. Qed.

Lemma ran_domUrangeS : [set x | existsb y, r x y] \subset domUrange r.
Proof. exact: subsetUl. Qed.

Lemma in_dom_in_domUrange x y : r x y -> x \in domUrange r.
Proof.
move=> rxy. 
by apply: (subsetP ran_domUrangeS); rewrite inE; apply/existsP; exists y.
Qed.

Lemma in_ran_in_domUrange x y : r x y -> y \in domUrange r.
Proof.
move=> rxy.
by apply: (subsetP dom_domUrangeS); rewrite inE; apply/existsP; exists x.
Qed.

Lemma relU_domUrange r1 r2 : 
  domUrange (relU r1 r2) = (domUrange r1) :|: (domUrange r2).
Proof.
apply/eqP; rewrite eqEsubset; apply/andP; split; apply/subsetP=> x; rewrite !inE.
  case/orP; case/existsP=> y /=; case/orP=> hxy.
  - suff -> : existsb y0, r1 x y0 by [].
    by apply/existsP; exists y.
  - suff -> : existsb y0, r2 x y0 by rewrite !orbT.
    by apply/existsP; exists y.
  - suff -> : existsb y0, r1 y0 x by rewrite !orbT.
    by apply/existsP; exists y.
  - suff -> : existsb y0, r2 y0 x by rewrite !orbT.
    by apply/existsP; exists y.
case/or3P.
  - case/orP;  case/existsP=> y /= hxy.
      suff -> : existsb y0, r1 x y0 || r2 x y0 by [].
      by apply/existsP; exists y; rewrite hxy.
    suff -> : existsb y0, r1 y0 x || r2 y0 x by rewrite orbT.
    by apply/existsP; exists y; rewrite hxy.
  -  case/existsP=> y /= hxy.
     suff -> : existsb y0, r1 x y0 || r2 x y0 by [].
     by apply/existsP; exists y; rewrite hxy orbT.
  -  case/existsP=> y /= hxy.
     suff -> : existsb y0, r1 y0 x || r2 y0 x by rewrite orbT.
     by apply/existsP; exists y; rewrite hxy orbT.
Qed.
    

Lemma subrel_trans r1 r2 r3 : 
  subrel r1 r2 -> subrel r2 r3 -> subrel r1 r3.
Proof. by move=> sr12 sr23 x y r1xy; apply: sr23; apply: sr12. Qed.

Lemma restr_transitive  (s : {set F}) : 
  transitive r -> transitive [rel x y \in s | r x y].
Proof.
move=>  rtrans x y z /=; case/andP; case/andP=> sy sx ryx.
by case/andP; case/andP=> _ sz rxz; rewrite sy sz (rtrans _ _ _ ryx).
Qed.

Lemma restr_irreflexive (s : {set F}) : 
  irreflexive r -> irreflexive [rel x y \in s | r x y].
Proof. by move=> rirr x /=; rewrite rirr andbF. Qed.

Lemma subrel_cycle (r1 r2 : rel F) : forall s,
  subrel r1 r2 -> cycle r1 s -> cycle r2 s.
Proof. case=> [| x s] //= sr12; exact: sub_path. Qed.

Lemma trans_path s x y :
  transitive r -> path r x (rcons s y) -> r x y.
Proof.
move=> rtrans; elim: s x y => [s1 y| s1 s hi x y] /=; first by rewrite andbT.
case/andP=> rxy hp; apply: rtrans rxy _; exact: hi.
Qed.

Lemma trans_cycle s x : transitive r -> cycle r (x :: s) -> r x x.
Proof. by move=> rtrans /=; move/(trans_path rtrans). Qed.


Lemma asymmetricP (e : rel F) :
  asymmetric e <-> (irreflexive e /\ antisymmetric e).
Proof.
split.
  move=> ae; split; last by move=> x y; move: (ae x y)=> /negPf->.
  by move=> x; move: (ae x x); rewrite andbb=> /negPf.
by case=> ie ae x y; apply/negP=> he; move: he (he)=> /ae ->; rewrite andbb ie.
Qed.


Lemma asym_trans_acyclic (e : rel F) (s : {set F}) 
  (te : {in s & &, transitive e}) (ae : {in s &, asymmetric e}) : acyclic_on s e.
Proof.
move=> [|x c] // _; rewrite subset_all /= => /andP [sx].
elim: c; first by rewrite /= andbT -[e _ _]andbb=> _; apply/negP; rewrite ae.
move=> y c ihc /= /andP [sy sc] /andP [exy pycx].
apply: ihc=> //; case: c pycx sc => /= [|z c hyz hz].
  by rewrite !andbT=> eyx _; apply: te eyx.
by case/andP: hyz=> eyz ->; case/andP: hz=> sz sc; rewrite andbT; apply: te eyz.
Qed.

Lemma acyclic_asym (e : rel F) (s : {set F}) 
   (ae : acyclic_on s e) : {in s &, asymmetric e}.
Proof.
move=> x y hx hy /=; apply/negP=> /andP [exy eyx].
have: cycle e [::x; y] by rewrite /= exy eyx.
by apply: ae=> //; rewrite subset_all /= hx hy.
Qed.

Lemma acyclic_restr s r1 : acyclic_on s r1 <-> acyclic [rel x y \in s | r1 x y].
Proof.
split=> h [| x p] //= _.
  move=> _ hp; apply: (h (x :: p)) => //=.
    move: hp; rewrite path_rcons /=; case/and3P=> hp; case/andP=> _ sx _.
    elim: p x sx hp {h} => [| y p ihp] x sx //=.
      by move=> _; apply/subsetP=> u; rewrite inE /=; move/eqP->.
    (do 3 case/andP)=> _ sy r1xy hp.
    have hind : y :: p \subset s by exact: ihp.
    apply/subsetP=> u; rewrite inE /=; case/orP; first by move/eqP->.
    by apply: (subsetP hind).
  by apply: (sub_path _ hp) => u v /=; do 2 case/andP.
move=> sp hp; apply: (h (x :: p)) => //=.
  apply: subset_trans sp _; exact: subsetT.
apply/(pathP x) => i si /=; move/(pathP x): hp; move/(_ i si)->.
rewrite /= !(subsetP sp) // -(mem_rcons p x) ?mem_nth //. 
case: i si => [|i] /= hs; first by rewrite mem_rcons mem_head.
rewrite mem_nth //; exact: ltn_trans hs.
Qed.

Lemma path_connect_irr p x y r1 : path (connect_irr r1) x (rcons p y) -> 
   exists p', path r1 x (rcons p' y).
Proof.
elim: p x y => [|z p ihp] x y //=.
  rewrite andbT; case/andP=> nexy; case/connectP=> p hp hlast.
  case/lastP: p hp hlast => [|p z] /= hp hlast.
    by move: nexy; rewrite hlast eqxx.
  by move: hlast; rewrite last_rcons => e; exists p; rewrite e.
(do 2 case/andP) => nexy; case/connectP=> p1 hp1 ez; case/ihp=> p2 hp2.
by exists (p1 ++ p2); rewrite rcons_cat path_cat hp1 -ez.
Qed.

Lemma acyclic_connect_irr r1 : acyclic r1 -> acyclic (connect_irr r1).
Proof.
move=> ha [| x l] //= sl _; rewrite /cycle.
case/path_connect_irr => l' hp'; apply: (ha (rcons l' x)); rewrite ?size_rcons //.
  by apply/subsetP.
by rewrite -rot1_cons cycle_rot /=.
Qed.
 
Lemma connect_irr_trans r1 : acyclic r1 -> transitive (connect_irr r1).
Proof.
move=> ac1 y x z; case/andP=> nexy cxy; case/andP=> neyz cyz. 
rewrite /connect_irr /=; case exz: (x == z); last exact: connect_trans cyz.
case/connectP: cyz=> p1 hp1 lp1; move: cxy; rewrite (eqP exz).
case/connectP=> p2 hp2 lp2.
case/lastP: p2 hp2 lp2 => [|p2 z2] /= hp2; rewrite ?last_rcons=> lp2.
  by move: neyz; rewrite lp2 eqxx.
(* this is a bug in 1.3pl1 *)
(* elim: (ac1 (y :: (p1 ++ p2))) => //; first by apply/subsetP.*)
suff: False by [].
apply: (ac1 (y :: (p1 ++ p2))) => //; first by apply/subsetP.
by rewrite /= rcons_cat path_cat hp1 -lp1 lp2.
Qed.

Lemma connect_irr_irr r1 : irreflexive (connect_irr r1).
Proof. by move=> x; rewrite /connect_irr /= eqxx. Qed.

Lemma connect_irr_asym r1 : acyclic r1 -> asymmetric (connect_irr r1).
Proof. move/acyclic_connect_irr; move/acyclic_asym=> h x y; exact: h. Qed.

End ExtraRelLemmas.

Section ExtraNatLemmas.

(* used in pigeon hole *)
Lemma wlog_leq : forall P, (forall a b, P b a -> P a b)
  -> (forall a b, a <= b -> P a b) -> forall a b, P a b.
Proof.
move=> P sP hP a b; wlog: a b / a <= b=> [hwlog|]; last exact: hP.
by case: (leqP a b); [|move/ltnW]; move/hwlog=> //; move/sP.
Qed.

(* the strict version, not used here *)
Lemma wlog_ltn : forall P, (forall a, P a a) -> (forall a b, (P b a -> P a b))
  -> (forall a b, a < b -> P a b) -> forall a b, P a b.
Proof.
move=> P rP sP hP; apply: wlog_leq=> // a b.
by rewrite leq_eqVlt; case: (altP (_ =P _))=> [->|] //= _ lab; apply: hP.
Qed.

End ExtraNatLemmas.


Section PigeonHole.

Variable F : finType.

(* A convenient version of the pigeon hole lemma, suggested by C. Cohen.*)
Lemma pigeon_hole (aT : finType) (f : aT -> F) : #|F| < #|aT| ->
  (exists x : aT, exists2 y : aT, x != y & f x = f y).
Proof.
case: (altP (injectiveP f)); last by move/injectivePn.
by move/(card_imset predT)<-; rewrite ltnNge subset_leq_card // subset_predT.
Qed.

Lemma pigeon_hole_nat (f : nat -> F) :
  exists i, exists2 d, i + d < #|F| & f i = f (i + d.+1).
Proof.
pose g (n : 'I_#|F|.+1) := f n.
case: (pigeon_hole g)=> [|[i hi] [[j hj]]]; first by rewrite ?card_ord.
rewrite /g /= -val_eqE /=; move: i j hi hj; apply: wlog_leq.
* by move=> i j hwlog hi hj hij fij; apply: hwlog=> //; rewrite eq_sym.
move=> i [|j hij hi hj nij fij] //=; first by rewrite leqn0=> /eqP-> //.
by exists i; exists (j - i); rewrite ?addnS ?subnKC // -ltnS ltn_neqAle nij.
Qed.

Definition pigeon_cycle (e : rel F) (s : pred F) (x0 : F) (sx0 : x0 \in s)
  (he : forall x, x \in s -> exists2 y, y \in s & e x y) :
  exists2 t : seq F, (size t > 0) && (t \subset s) & cycle e t.
Proof.
have {he} he x (hx : x \in s) : {y | (y \in s) && (e x y)}.
  by apply: sigW; case: (he _ hx)=> y hy exy; exists y; rewrite hy.
pose f a := if (a \in s) =P true is ReflectT h then val (he _ h) else a.
have fP a (ha : a \in s) : (f a \in s) * (e a (f a)).
  by rewrite /f; case: eqP=> // p0; case: he f=> ? /=; case/andP=> [].
case: (pigeon_hole_nat (fun k => iter k f x0))=> [i [d _]].
rewrite addnC iter_add; set x := iter _ _ _=> hx.
have sx: x \in s by rewrite /x; elim: i {x hx}=> [|i ihi] //=; rewrite fP.
exists (traject f x d.+1); rewrite ?andTb.
  elim: d.+1 x sx {hx}=> {d} [|d hd] x sx //=; apply/subsetP=> // y.
  rewrite in_cons; case/orP; first by move/eqP->.
  by apply/subsetP; rewrite hd // fP.
rewrite /= {3}hx -addn1 iter_add /= -trajectSr.
by elim: d.+1 x sx {hx}=> {d} [|d hd] x sx //=; rewrite ?(fP, hd).
Qed.

Definition pigeon_cycle_rev (e : rel F) (s : pred F) (x0 : F) (sx0 : x0 \in s)
  (he : forall x, x \in s -> exists2 y, y \in s & e y x) :
  exists2 t : seq F, (size t > 0) && (t \subset s) & cycle e t.
Proof.
case: (@pigeon_cycle (fun a => e^~ a) s x0)=> // [] [] // x t. 
case/andP=> _ hs ct; exists (x :: rev t)=> //.
  by apply/subsetP=> a; rewrite in_cons mem_rev -in_cons; apply/subsetP.
by move: ct; rewrite /= -path_rev last_rcons belast_rcons rev_cons.
Qed.


End PigeonHole.

Section TopologicalSorting.

Variable F : finType.

Fixpoint topo_rec (e : rel F) (s : {set F}) (n : nat) :=
  if n is n'.+1 then
    if [pick x \in s | forallb y, (y \in s) ==> ~~ e y x] is Some x
      then let e' := topo_rec e (s :\ x) n' in
        fun a b => if (a == x) && (b \in (s :\ x)) then true else e' a b
      else e
    else e.


Lemma topo_rec_id (e: rel F) (s : {set F}) (n : nat) 
  (x y : F) (hxy : (x \notin s) || (y \notin s)) : topo_rec e s n x y = e x y.
Proof.
elim: n s hxy => [|n ihn] s hxy //=; case: pickP=> // x0.
case: (altP (x =P x0))=> xx0 /=.
  rewrite -xx0; case: hxy=> /orP [/negPf-> //|sy].
  by rewrite ihn ?in_setD1 ?eqxx (negPf sy) ?andbF.
move=> /andP [sx0 /forallP he]; rewrite ihn // ?in_setD1 xx0 /=.
by case: hxy=> /orP [] /negPf->; rewrite ?andbF ?orbT.
Qed.

Lemma topo_rec_subrel (e : rel F) (s : {set F}) (n : nat):
   subrel e (topo_rec e s n).
Proof.
move=> x y /=; elim: n e s => //= n ihn e s exy.
case: pickP=> //= x0 /=; case/andP=> hx0; move/forallP=> hx1.
by case: ifP=> // ?; apply: ihn.
Qed.

Lemma topo_rec_asym (e : rel F) (s : {set F}) (s' : {set F})
  (ae : {in s' &, asymmetric e}) (n : nat) :
  {in s' &, asymmetric (topo_rec e s n)}.
Proof.
elim: n e s ae=> [|n ihn] e s ae /= x y s'x s'y; first exact: ae.
case: pickP; last by move=> *; apply: ae.
move=> x0 /andP [hx0 /forallP hx1].
have {hx1} hx1 : forall x : F, (x \in s) -> ~~ e x x0.
  by move=> x1; apply/implyP.
rewrite !in_setD1 /=.
case: (altP (x =P x0)) s'x=> [->|hx] s'x;
case: (altP (y =P x0)) s'y=> [->|hy] s'y //=; do ?exact: ihn;
by do [ case: ifP=> /= sy; last exact: ihn;
        rewrite topo_rec_id ?andbT ?hx1 // ?in_setD1 eqxx orbT].
Qed.

Lemma topo_rec_total (e : rel F) (s : {set F}) :
  acyclic_on s e ->  {in s &, total_irr (topo_rec e s #|s|)}.
Proof.
move hs: #|s|=> n.
elim: n s e hs => [|n ihn] s e hs ae /=.
  by rewrite [s]cards0_eq // => x; rewrite in_set0.
move=> x y sx sy /=.
case: pickP=>[|he]; last first.
  suff: False by [].
  case: (@pigeon_cycle_rev _ e (mem s) x); rewrite ?mem_enum //.
    move=> a; rewrite ?mem_enum=> ha; move: (he a); rewrite ha /= => /negbT.
    rewrite negb_forall=> /existsP [b]; rewrite negb_imply negbK.
    by case/andP=> [sb eba]; exists b; rewrite ?mem_enum ?sb.
  by move=> t /andP [sizet] //; apply: ae.
move=> x0 /andP [sx0 /forallP hx0] /=.
case: (altP (x =P x0)) sx=> [->|xx0] sx;
case: (altP (y =P x0)) sy=> [->|yx0] sy //=.
* by rewrite eqxx.
* by rewrite ?in_setD1 yx0 sy /= orbT.
* by rewrite ?in_setD1 xx0 sx /= !orbT.
apply: ihn.
* by move: hs; rewrite (cardsD1 x0) sx0 add1n=> [] [].
* move=> c hc /subsetP sc; apply: ae=> //; apply/subsetP=> a /sc.
  by rewrite in_setD1=> /andP [].
* by rewrite in_setD1 (negPf xx0).
by rewrite in_setD1 (negPf yx0).
Qed.

Lemma topo_rec_trans (e : rel F) (s : {set F}) (n : nat)
  (te : {in s & &, transitive e}) : {in s & &, transitive (topo_rec e s n)}.
Proof.
elim: n s e te => [|n ihn] s e te y x z sx sy sz; first exact: te.
rewrite /=; case: pickP; last by move=> ?; apply: te.
move=> x0 /andP [sx0 /forallP he] /=.
have {he} he x' : x' \in s -> ~~ e x' x0 by apply/implyP.
case: (altP (x =P x0))=> xx0 /=; case: (altP (y =P x0))=> yx0 /=;
  case: (altP (z =P x0))=> zx0 /=; 
    rewrite ?(xx0, yx0, zx0, sx, sy, sz, in_setD1, eqxx) //=.
* by rewrite ?topo_rec_id ?(in_setD1, eqxx, orbT) // (negPf (he _ _)).
* by rewrite topo_rec_id ?(in_setD1, eqxx, orbT) // (negPf (he _ _)).
* by move=> _; rewrite topo_rec_id ?(in_setD1, eqxx, orbT) // (negPf (he _ _)).
apply: ihn; rewrite ?hs' ?(in_setD1, xx0, yx0, zx0) //.
by move=> a b c /=; rewrite ?in_setD1; do ?case/andP=> _ ?; apply: te.
Qed.

Lemma topo_rec_acyclic (e : rel F) (s : {set F}) (n :nat)
  (ae : {in s &, asymmetric e}) (te : {in s & &, transitive e}) :
  acyclic_on s (topo_rec e s n).
Proof.
by apply: asym_trans_acyclic; [apply: topo_rec_trans|apply: topo_rec_asym].
Qed.


Lemma topo_rec_domUrange (e : rel F) (s1 s2 : {set F}) (n :nat) :
  domUrange e \subset s1 -> domUrange (topo_rec e s2 n) \subset s1 :|: s2.
Proof.
move: e s2; elim: n => [|n ihn] e s2 hs /=.
  by apply: subsetU; rewrite hs.
case: pickP=> // x0; last  by apply: subsetU; rewrite hs.
case/andP => sx0 hx0.
have {hx0} hx0 y: y \in s2 -> ~~e y x0.
  by move=> sy; move/forallP: hx0; move/(_ y); move/implyP; apply.
apply/subsetP=> y; rewrite !inE /=; case/orP; case/existsP=> z.
  case: (altP (y =P x0)) sx0 hx0 => [<-|xx0] sy hy /=; first by rewrite sy orbT.
  by move/in_dom_in_domUrange; move/(subsetP (ihn _ _ hs)); rewrite !inE xx0.
case: ifP; first by case/and3P=> _ _ ->; rewrite !orbT.
move=> _; move/in_ran_in_domUrange; move/(subsetP (ihn _ _ hs)); rewrite !inE.
by case/orP; [move-> | case/andP=> _ ->; rewrite orbT].
Qed.

Definition toposort (e : rel F)(s : {set F}) := topo_rec e s #|s|.

End TopologicalSorting.
