(* Equivalence of a  litteral implementation of Lamport Sequential Consistency*)
(* with the corresponding instance of the framework *)
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice.
Require Import  path fintype finset fingraph.
Require Import extrarel wmm.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Section FixedSize.

Variable n : nat.

Notation "'Word" := (Word n).
Notation "'Address" := (Address n).
Notation "'Value" := (Value n).
Notation "'Proc" := (Proc n).
Notation "'Eiid" := (Eiid n).
Notation "'Location" := (Location n).
Notation "'Event" := (Event n).
Notation "'Event_struct" := (Event_struct n).
Notation "'Action" := (Action n).
Notation "'Execution_witness" := (Execution_witness n).

Section SCArchitectures.


(* A sequentially consistent architecture imposes the value of ppo, inter, 
intra, dp but imposes no restriction on abc *)

Definition po_iico_ppo (E : 'Event_struct) := (po_iico_of_ES E).

Lemma po_iico_ppo_valid : forall E, subrel (po_iico_ppo E) (po_iico_of_ES E).
Proof. by move=> E. Qed.

Lemma rf_grf_valid : forall X : 'Execution_witness, 
  subrel (rf_of_EW X) (rf_of_EW X).
Proof. by move=> *. Qed.


(* At the end of the section, abc and ab_events are abstracted in the definition
of mkSCArchi *)
Variable abc :  'Event_struct -> 'Execution_witness -> rel 'Event.
Hypothesis ab_evts : forall E X, subrel (abc E X) [rel x y \in (events_of_ES E)].

Definition mkSCArchi := mkArchi rf_grf_valid
  po_iico_ppo_valid ab_evts.

End SCArchitectures.


Section SequentiallyConsistentsEventStructures.

(* E is sequentially consistant with respect to so *)

Definition previous_writes (E : 'Event_struct)(r : rel 'Event)(er : 'Event) :=
  [set ew \in writes E | (r ew er) && (loc_of_Event ew == loc_of_Event er)].

Definition maximal_elements (xs : {set 'Event})(r : rel 'Event) :=
  [set x \in xs | forallb y : 'Event, ((y \in xs) && r x y) ==> (x == y)].

Lemma maximal_elementsP : forall (xs : {set 'Event})(r : rel 'Event)(x : 'Event),
  reflect ((x \in xs) /\ (forall y : 'Event, y \in xs -> r x y -> x = y))
    (x \in maximal_elements xs r).
Proof.
move=> xs r x; apply: (iffP idP) => /=; rewrite inE /=.
  case/andP=> xsx; move/forallP=> /= hmax; split=> // y xsy rxy; have := (hmax y).
  by rewrite xsy rxy /=; move/eqP.
case=> -> hmax /=; apply/forallP=> /= y; apply/implyP; case/andP=> ? ?; apply/eqP.
exact: hmax.
Qed.


Definition so_rfm (E : 'Event_struct) (so : rel 'Event)(ew er : 'Event) :=
    (rfmaps  ew er)
    && 
    (ew \in maximal_elements (previous_writes E so er) so).

Lemma so_rfm_soS : forall E so,  subrel (so_rfm E so) so.
Proof.
move=> E so e1 e2; case/andP=> _; case/maximal_elementsP; rewrite inE.
by case/and3P.
Qed.

Lemma so_rfm_inj : forall E (so : rel 'Event) w1 w2 x, so w1 w2 ->
  so_rfm E so w1 x -> so_rfm E so w2 x -> w1 = w2.
Proof.
move=> E so w1 w2 x sow12.
case/andP=> _; case/maximal_elementsP=> pw_w1 w1max.
case/andP=> _; case/maximal_elementsP=> pw_w2 w2max.
by rewrite (w1max _ pw_w2).
Qed.

Lemma so_rfm_isWl : forall E so e1 e2, so_rfm E so e1 e2 -> dir_of_Event e1 = W.
Proof. by move=> E so e1 e2;case/andP; case/and3P; move/eqP. Qed.

Lemma so_rfm_isRr : forall E so e1 e2, so_rfm E so e1 e2 -> dir_of_Event e2 = R.
Proof. by move=> E so e1 e2; case/andP; case/and3P=> _ ; move/eqP. Qed.

Lemma so_rfm_loc : forall E so e1 e2, 
  so_rfm E so e1 e2 -> loc_of_Event e1 = loc_of_Event e2.
Proof. by move=> E so e1 e2; case/andP; case/and3P=> _ _; move/eqP. Qed.

Lemma so_rfm_max: forall E so e1 e2, so_rfm E so e1 e2 -> 
  (forall y : 'Event, y \in (previous_writes E so e2) -> so e1 y -> e1 = y).
Proof. by move=> E so e1 e2; case/andP=> _; case/maximal_elementsP. Qed.

Definition seq_exec (E : 'Event_struct) (so : rel 'Event) : Prop := 
  [/\ {in (events_of_ES E) &, total_irr so},
      irreflexive so,
      (domUrange so) \subset (events_of_ES E),
      transitive so &
      subrel (po_iico_of_ES E) so] /\
      (forall er : 'Event, er \in events_of_ES E ->
        exists ew : 'Event, ew \in events_of_ES E /\ so_rfm E so ew er).


Definition so_ws (so : rel 'Event)(e1 e2 : 'Event) :=
  (so e1 e2) && writes_to_same_loc (domUrange so) e1 e2.

Lemma so_ws_loc : forall (so : rel 'Event) e1 e2, so_ws so e1 e2 ->
  loc_of_Event e1 = loc_of_Event e2.
Proof. by move=> so e1 e2; case/and5P=> _ _ _ _; move/eqP. Qed.

Lemma so_ws_isWl : forall (so : rel 'Event) e1 e2, so_ws so e1 e2 -> dir_of_Event e1 = W.
Proof. by move=> so e1 e2; case/and5P=> _ _; move/eqP. Qed.

Lemma so_ws_isWr : forall (so : rel 'Event) e1 e2, so_ws so e1 e2 -> dir_of_Event e2 = W.
Proof. by move=> so e1 e2; case/and5P=> _ _ _; move/eqP. Qed.

Lemma so_ws_soS : forall so, subrel (so_ws so) so.
Proof. by move=> so e1 e2; case/andP. Qed.

Lemma so_ws_trans : forall so, transitive so -> transitive (so_ws so).
Proof.
move=> so so_trans x y z; case/andP=> so_yx Wto_yx; case/andP=> so_xz Wto_xz.
by rewrite /so_ws (so_trans _ _ _ so_yx) // (W_to_same_loc_trans Wto_yx).
Qed.

Lemma so_ws_irrefl : forall so, irreflexive so -> irreflexive (so_ws so).
Proof. by move=> so so_irrefl x; rewrite /so_ws so_irrefl. Qed.

End SequentiallyConsistentsEventStructures.

Section SCEquivalence.

Definition abc_false (E : 'Event_struct)(X : 'Execution_witness) : rel 'Event :=
  fun x y => false.

Lemma abc_falseP : forall E X, subrel (abc_false E X) [rel x y \in (events_of_ES E)].
Proof. by []. Qed.

Local Notation "'A" := (mkSCArchi abc_falseP).


Lemma native_sc_is_sc : forall E rfm wsn so, 
  wf_ES E -> seq_exec E so -> so_rfm E so =2 rfm -> so_ws so =2 wsn ->
    (exists X, 
      [/\ wf_EW E X, 
        valid_execution 'A E X, (rf_of_EW X) =2 rfm & (ws_of_EW X) =2 wsn]).
Proof.
move=> E rfm wsn so wfE [[hlin hirr hsub htrans hsubrel] hinit hso_rfm hso_ws].
  have fr_sub : subrel (fr (mkew (so_ws so) (so_rfm E so))) so.
    move=> e1 e2; case/existsP=> /= y; case/andP=>  rfye1 wsye2.
    have ev_e1 : e1 \in events_of_ES E.
      by apply: (subsetP hsub); apply: (in_ran_in_domUrange (so_rfm_soS rfye1)).
    have ev_e2 : e2 \in events_of_ES E.
      by apply: (subsetP hsub); apply: (in_ran_in_domUrange (so_ws_soS wsye2)).
    move: (hlin _ _ ev_e1 ev_e2); case/orP=> // e12.
        move: (so_ws_isWr wsye2); rewrite -(eqP e12).
        by rewrite (so_rfm_isRr rfye1).
    case/orP: e12=> // soe21; have soye2 := so_ws_soS wsye2.
    suff eye2: y = e2 by move: soye2; rewrite eye2 (hirr e2).
    apply: (so_rfm_max rfye1); rewrite // inE /= soe21 inE /= ev_e2.
    by rewrite (so_ws_isWr wsye2) -(so_ws_loc wsye2) -(so_rfm_loc rfye1) !eqxx.
exists (mkew  (so_ws so) (so_rfm E so)); split.
- split.
  * split; split => l /=.
    + apply: restr_transitive; exact: so_ws_trans.
    + apply: restr_irreflexive; exact: so_ws_irrefl.
    + move=> e1 e2 wle1 wle2; rewrite /so_ws /writes_to_same_loc /=.
      case/W_toSP: wle1 => Ee1 -> ->; case/W_toSP: wle2 => Ee2 -> ->. 
      rewrite !eqxx !andbT.
      have -> : 
        [&& so e1 e2, e1 \in domUrange so & e2 \in domUrange so] = so e1 e2.
         case soe12 : (so e1 e2) => //=; rewrite (in_dom_in_domUrange soe12).
         by rewrite (in_ran_in_domUrange soe12).
      have -> : 
        [&& so e2 e1, e2 \in domUrange so & e1 \in domUrange so] = so e2 e1.
        case soe21 : (so e2 e1) => //=; rewrite (in_dom_in_domUrange soe21).
        by rewrite (in_ran_in_domUrange soe21).
      exact: hlin.
   + move=> e2;rewrite /so_ws /=; case/andP => sole2 _; apply: (subsetP hsub).
    by rewrite (in_dom_in_domUrange sole2).
   + move=> e2;rewrite /so_ws /=; case/andP => sole2 _; apply: (subsetP hsub).
    by rewrite (in_ran_in_domUrange sole2).
   + move=> e2; rewrite /wsmaps; case/andP=> _; rewrite /writes_to_same_loc /=.
    by case/andP.
  * split.
   + by move=> e Ee; case: (hinit e Ee) => e2 [Ee2 soe2]; exists e2.
   + move=> e1 e2 e3 /= h1 h2.
     case/andP:(h1)=> [] rfh1; case/maximal_elementsP=> [pw1 max1].
     move: (pw1); rewrite inE /=; case/and3P=> we2 so21; move/eqP=> loc21.
     case/andP:(h2)=> [] rf2; case/maximal_elementsP=> [pw2 max2].
     move: (pw2); rewrite inE /=; case/and3P=> we1 so31; move/eqP=> loc31.
     move: we1 we2; rewrite !inE /=; case/andP=> E1 _; case/andP=> E2 _.
     move: (hlin _ _ E1 E2); case/orP; first by move/eqP.
     case/orP=> hso; first by apply: sym_eq; apply: max2.
     exact: max1.
   + move=> e1 e2 /=; case/andP=> _; case/maximal_elementsP; rewrite inE; case/and3P.
     by rewrite inE; case/andP.
   + move=> e1 e2 /=; case/andP=> _;  case/maximal_elementsP; rewrite inE. 
     by case/and3P => _ h _ _; apply: (subsetP hsub); apply: (in_ran_in_domUrange h).
   + by move=> e1 e2 /=; case/andP.    
- split.
  * case=> // x s _  _ hc.
    suff soxx : so x x by move: (hirr x); rewrite soxx.
    suff hi : subrel (relU (com (mkew (so_ws so) (so_rfm E so))) (po_loc E)) so.
      have {hc} hc : cycle so (x :: s) by apply: (subrel_cycle hi).
      exact: (trans_cycle htrans hc).
    move=> e1 e2; case/orP; last by case/andP=> _; move/hsubrel.
    case/orP; last by case/andP. 
    case/orP; [exact: so_rfm_soS | exact: fr_sub].
  * case=> // x s _ _ hc.
    suff soxx : so x x by move: (hirr x); rewrite soxx.
    suff hi : subrel (relU (rf_of_EW (mkew (so_ws so) (so_rfm E so))) (dp_of_ES E)) so.
      have {hc} hc : cycle so (x :: s) by apply: (subrel_cycle hi).
      exact: (trans_cycle htrans hc).
    move=> e1 e2; case/orP => /=.
    + by case/andP=> _; case/maximal_elementsP; rewrite inE; case/and3P.
    case: wfE=> _ [] => hs _ _; move/hs; exact: hsubrel.
  * case=> // x s _ _ hc.
    suff soxx : so x x by move: (hirr x); rewrite soxx.
    suff hi : subrel (ghb 'A E (mkew (so_ws so) (so_rfm E so))) so.
       have {hc} hc : cycle so (x :: s) by apply: (subrel_cycle hi).
       exact: (trans_cycle htrans hc).
    move=> e1 e2; case/orP => /=.
    (* we should prove more general things on grf... *)
      by case/andP=> _; rewrite inE /=; case/andP; rewrite inE /=; case/and3P.
    case/orP; last by move/hsubrel.
    by case/orP; [case/andP | move/fr_sub].
- exact: hso_rfm. 
- exact: hso_ws.
Qed.

Lemma sc_is_sc : forall E rfm wsn, 
  wf_ES E -> 
  ((exists so, [/\ seq_exec E so, so_rfm E so =2 rfm & so_ws so =2 wsn]) <->
    (exists X, 
      [/\ wf_EW E X, 
        valid_execution 'A E X, (rf_of_EW X) =2 rfm & (ws_of_EW X) =2 wsn])).
Proof.
move=> E rfm wsn wfE; split.
  case=> so [[[hlin hirr hsub htrans hsubrel] hinit] hso_rfm hso_ws].
  exact: (@native_sc_is_sc _ _ _ so). (* exact with @ _ _ _ so is way too long ...*)
case=> X [] [[[htrans hirr hlin]] [wevtl wevtr hwsmaps]] [] hrf1 hrf2 revtl revtr hrfmaps.
case => hacycl1 hacycl2 hacycl3 validX hws2.
set Ev := events_of_ES E. 
pose r1 :=  connect_irr (ghb 'A E X).
pose r2 := [rel x y \in Ev | r1 x y].
pose r := toposort r2 Ev.
have r21_sub : subrel r2 r1 by move=> x y; case/andP.
have ac_r1 : acyclic r1 by apply: acyclic_connect_irr.
have trans_r1 : transitive r1 by apply: connect_irr_trans.
have r2_asym : {in Ev &, asymmetric r2}.
  move=> x y Ex Ey /=.
  rewrite Ex Ey /=. rewrite !andbA.
  rewrite -[(x != y) && connect _ x y]/(r1 x y).
  rewrite -andbA -[(y != x) && connect _ y x]/(r1 y x).
  exact: connect_irr_asym.
have trans_r2 : {in Ev & &, transitive r2}.
  move=> x y z Ex Ey Ez /=; rewrite Ex Ey Ez /= => r1yx r1xz.
  exact: trans_r1 r1yx _. 
have ac_r : acyclic_on Ev r by exact: topo_rec_acyclic => //.
have rdU_r : domUrange r \subset Ev.
  rewrite -(setUid Ev); apply: topo_rec_domUrange.
  by apply/subsetP=> x; rewrite !inE /=; case/orP; case/existsP=> y; do 2 case/andP.
have tot_r : {in Ev &, total_irr r}. 
  apply: topo_rec_total; exact: asym_trans_acyclic.
have trans_r : {in Ev & &, transitive r} by apply: topo_rec_trans.
have irr_r : {in Ev, irreflexive r}.
  (* this should be a generic lemma *)
  move=> x Ex /=.
  suff: ~ cycle r [:: x] by rewrite /= andbT; move/negP; move/negPf.
  apply: ac_r; first by []. 
  by apply/subsetP=> y; rewrite inE; move/eqP->.
have asym_r : {in Ev &, asymmetric r} by apply: acyclic_asym.
exists r => {r21_sub r2_asym trans_r2}; split; last 2 first.
- move=> e1 e2; rewrite -validX; apply/idP/idP => hxy.
  * move: (hxy); rewrite /so_rfm. case/andP=> hrfmaps12 max2.  
    case/maximal_elementsP: max2; rewrite inE; case/and3P => We1 r12 hloc hmax1.
    have Ee1 : e1 \in events_of_ES E.
      apply: (subsetP rdU_r); apply: (subsetP (ran_domUrangeS _)); rewrite inE /=.
      by apply/existsP; exists e2.
    have Ee2 : e2 \in events_of_ES E.
      apply: (subsetP rdU_r); apply: (subsetP (dom_domUrangeS _)); rewrite inE /=.
      by apply/existsP; exists e1.
    case: (hrf1 _ Ee2) => w hrfwe2.
    have hloc1 : e1 \in (writes_toS E (loc_of_Event e1)).
      by rewrite inE /= Ee1 eqxx andbT; move: We1; rewrite inE /=; case/andP.
    have hlocw : w \in (writes_toS E (loc_of_Event e1)).
      rewrite inE (revtl  _ _ hrfwe2) /=.
      move: (hrfmaps _ _ hrfwe2); rewrite /rfmaps; case/and3P=> -> _; move/eqP ->.
      by rewrite (eqP hloc) eqxx.
    move: (hlin _ _ _ hloc1 hlocw); case/orP; first by move/eqP->.  
    move/orP=> [he1w | hwe1].
    + suff -> : e1 = w by [].
      have rwe2 : r w e2.
        apply: topo_rec_subrel; rewrite /= Ee2 (wevtr _ _ he1w) /=. 
        case abs : (w == e2).
          move/hrfmaps: hrfwe2=> hwe2; move: (rfmaps_isWl hwe2).
          by rewrite (eqP abs) (rfmaps_isRr hwe2).
        by rewrite /=; apply: connect1; rewrite /ghb /= hrfwe2.
      case abs : (e1 == w); first by apply/eqP.
      have re1w : r e1 w.
        apply: topo_rec_subrel; rewrite /= Ee1 (wevtr _ _ he1w) abs /=. 
        by apply: connect1; rewrite /ghb /= he1w !orbT.
      apply: hmax1 => //; rewrite !inE /= (wevtr _ _ he1w) /=.
      case/and3P: (hwsmaps _ _ he1w) => hdirl hdirr hloce1w.
      by rewrite hdirr -(eqP  hloce1w) hloc andbT /=.
    + have fre2e1 : fr X e2 e1.
        by rewrite /fr /=; apply/existsP; exists w; rewrite hrfwe2.
      suff : r e2 e2 by rewrite (irr_r _ Ee2).
      case e12 : (e1 == e2); first by rewrite -{1}(eqP e12).
      apply: trans_r r12 => //; apply: topo_rec_subrel; rewrite /= Ee2 Ee1.
      by rewrite connect1  /ghb /= ?fre2e1 ?orbT // eq_sym e12.
  * rewrite /so_rfm /= (hrfmaps _ _ hxy) /=.
    apply/maximal_elementsP. rewrite inE.
    case/and3P: (hrfmaps _ _ hxy) => dir1 dir2 hloc12.
    have ne12 : e1 != e2.
      by apply/eqP=> e12; move: dir1; rewrite e12 (eqP dir2); move/eqP.
    have E1 : e1 \in Ev by exact: (revtl _ _ hxy).
    have E2 : e2 \in Ev by exact: (revtr _ _ hxy).
    rewrite inE /= dir1 hloc12 (revtl _ _ hxy)  /= andbT; split.
      by apply: topo_rec_subrel; rewrite /= E1 E2 ne12 connect1 //= /ghb /= hxy.
    move=> y; rewrite inE /=; case/and3P; rewrite inE /=; case/andP => Ey diry.
    move=> htopye2 hlocye2 htope1y.
    have We1 : e1 \in writes_toS E (loc_of_Event e1).
      by rewrite !inE eqxx dir1 !andbT.
    have Wy : y \in writes_toS E (loc_of_Event e1).
      by rewrite (eqP hloc12) -(eqP hlocye2) inE Ey  eqxx diry. 
    move: (hlin (loc_of_Event e1) _ _ We1 Wy). 
    case/orP; first by move/eqP->. 
    have ne2y: e2 != y.
      by apply/eqP=> e2y; move: dir2; rewrite e2y (eqP diry); move/eqP.
    have ne1y: y != e1.
      by case abs : (y == e1) => //; move: htope1y; rewrite (eqP abs) (irr_r _ E1).
    move/orP=> [he1w | hwe1].
    + have rwe2 : r e2 y.
      suff fre2y : fr X e2 y.
        apply: topo_rec_subrel; rewrite /= E2 Ey ne2y connect1 // /ghb /=.
        by rewrite fre2y !orbT.
      by rewrite /fr /=; apply/existsP; exists e1; rewrite hxy.
      by move: (asym_r e2 y E2 Ey); rewrite rwe2  htopye2.
    + have rwe1 : r y e1.
        apply: topo_rec_subrel; rewrite /= E1 Ey ne1y connect1 // /ghb /=.
        by rewrite hwe1 !orbT.
      by move: (asym_r e1 y E1 Ey); rewrite rwe1  htope1y.
- move=> e1 e2; rewrite -hws2; apply/idP/idP => hxy.
  * move: (hxy); rewrite /so_ws. case/and5P=> hr12 hudr hw1 hw2 hl.
    have We1 : e1 \in (writes_toS E (loc_of_Event e1)).
      rewrite inE /= hw1 eqxx /=.
      by move/in_dom_in_domUrange: hr12; move/(subsetP rdU_r)->.
    have We2 /= : e2 \in (writes_toS E (loc_of_Event e1)).
      rewrite inE /= hw2 (eqP hl) eqxx.
      by move/in_ran_in_domUrange: hr12; move/(subsetP rdU_r)->.
    move: (hlin (loc_of_Event e1) _ _ We1 We2).
    have E1 : e1 \in Ev by case/andP: hudr; move/(subsetP rdU_r).
    have E2 : e2 \in Ev by case/andP: hudr=> _; move/(subsetP rdU_r).
    have ne12 : e1 != e2.
      by case abs : (e1 == e2) => //; move: hr12; rewrite (eqP abs) (irr_r _ E2). 
   case/or3P=> [eq12 | //= | hw21].
      + by move: (irr_r _ E1); rewrite {2}(eqP eq12) hr12.
      + suff r21: r e2 e1 by move: (asym_r _ _ E1 E2); rewrite r21 hr12.
       apply: topo_rec_subrel; rewrite /= E1 E2 eq_sym ne12 connect1 // /ghb /=.
       by rewrite hw21 !orbT.
  * rewrite /so_ws.
    have E1 : e1 \in Ev by apply: (wevtl _  _ hxy).
    have E2 : e2 \in Ev by apply: (wevtr _  _ hxy).
    have ghb12 : ghb 'A E X e1 e2 by rewrite /ghb /= hxy !orbT.
    have ne12 : e1 != e2.
      case abs : (e1 == e2) => //.
      (* This is a bug in 1.3pl1, elim should work *)
      suff: False by [].
      apply: (hacycl3 [:: e1]) => //=; last by rewrite andbT {2}(eqP abs).
      by apply/subsetP.
    have r12 : r e1 e2.
     by apply: topo_rec_subrel; rewrite /= E1 E2 ne12 connect1.
    rewrite r12 /= /writes_to_same_loc /=.
    rewrite (in_dom_in_domUrange r12) (in_ran_in_domUrange r12) /=.
    by move/hwsmaps: hxy.
- split; last first.
- move=> er; case/hrf1=> ew hew; exists ew; split; first exact: (revtl _ _ hew).
  have hrfm : rfmaps ew er by exact: hrfmaps.
  have ghbrw :  ghb 'A E X ew er by rewrite /ghb /= hew.
  have newr : ew != er.
    case abs : (ew == er) => //.
 (* This is a bug in 1.3pl1,  elim should work*)
      suff: False by [].
      apply: (hacycl3 [:: ew]) => //=; last by rewrite andbT {2}(eqP abs).
    by apply/subsetP.
  rewrite /so_rfm hrfm /=; apply/maximal_elementsP; split.
    rewrite !inE /= (rfmaps_loc hrfm) eqxx (revtl _ _ hew).
    rewrite (rfmaps_isWl hrfm) eqxx /= andbT; apply: topo_rec_subrel.
    by rewrite /r2 /= newr (revtl _ _ hew) (revtr _ _ hew) connect1.
  move=> y; rewrite !inE; case/and3P; case/andP=> Ey dy ryer hloc rewy.
  have h1 : y \in writes_toS E (loc_of_Event er).
    by rewrite inE Ey dy.
  have h2 /= : ew \in writes_toS E (loc_of_Event er).
    rewrite inE (rfmaps_isWl hrfm) (rfmaps_loc hrfm) (revtl _ _ hew).
    by rewrite !eqxx.
  move: (hlin (loc_of_Event er) _ _ h2 h1); case/or3P; first by move/eqP.
  move=> wsewy.
  have frery : fr X er y. 
    by rewrite /fr /=; apply/existsP; exists ew; rewrite hew.
  have Er : er \in Ev.
    by apply: (subsetP rdU_r); apply: (in_ran_in_domUrange ryer).
  have nery : er != y.
    case abs: (er == y) => //.
      by have:= (irr_r _  Ey); rewrite -{2}(eqP abs) ryer.
  have rery : r er y.
    rewrite /r; apply: topo_rec_subrel; rewrite /r2 /= Er Ey nery.
    by rewrite connect1 // /ghb /= frery !orbT.
  by move: (asym_r _ _ Ey Er); rewrite rery ryer.
  - move=> hyew.
    have ghbyew : ghb 'A E X y ew by rewrite /ghb /= hyew !orbT.
    have neyew : y != ew.
      case abs : (y == ew) => //.
 (* This is a bug in 1.3pl1,  elim should work*)
      suff: False by [].
      apply: (hacycl3 [:: y]) => //=; last by rewrite andbT {2}(eqP abs).
      by apply/subsetP.
    have Eew : ew \in Ev by exact: wevtr _ _ hyew.
    have ryew : r y ew.
      by apply: topo_rec_subrel; rewrite /r2 /= Eew Ey neyew connect1.
    by move: (asym_r _ _ Eew Ey); rewrite ryew rewy.
split => //; last first.
- move=> x y hxy. apply: topo_rec_subrel.
  rewrite /r2 /=.
  have ghbxy : ghb 'A E X x y by rewrite /ghb /= /po_iico_ppo hxy !orbT.
  have nexy : x != y.
    case abs : (x == y) => //.
(* This is a bug in 1.3pl1,  elim should work*)
      suff: False by [].
      apply: (hacycl3 [:: x]) => //=; last by rewrite andbT {2}(eqP abs).
    by apply/subsetP.
  have Exy : (x \in Ev) && (y \in Ev).
    move: hxy; rewrite /po_iico_of_ES /=; case/orP.
      rewrite /po /=; case/andP; rewrite /same_proc_same_ES /=; case/andP.
      by case/andP=> -> ->.
    move=> hxy; case: wfE=> _ _; case=> _ hiico _.
    move: (in_dom_in_domUrange hxy); move/(subsetP hiico)->.
    by move: (in_ran_in_domUrange hxy);  move/(subsetP hiico)->.
  by rewrite Exy nexy /= connect1.
- move=> x y z ryx rxz.
  have Ex : x \in Ev.
    apply: (subsetP rdU_r); exact: (in_dom_in_domUrange rxz).
  have Ey : y \in Ev.
    apply: (subsetP rdU_r); exact: (in_dom_in_domUrange ryx).
  have Ez : z \in Ev.
    apply: (subsetP rdU_r); exact: (in_ran_in_domUrange rxz).
  exact: (trans_r _ _ _ Ex Ey Ez).
- move=> x; case rxx: (r x x) => //.
  have Ex : x \in Ev.
    apply: (subsetP rdU_r); exact: (in_dom_in_domUrange rxx).
  by have := (irr_r _ Ex); rewrite rxx.
Qed.

End SCEquivalence.
End FixedSize.