Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice.
Require Import  path fintype finset extrarel.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* An architecture is parameterized by some sizes and characterized by *)
(*: - the ppo relation
    - the inter and intra boolean flags
    - the abc relation
It is modelled by a record type: an archi instance is an inhabitant of *)
(*this type. One can then quantify over architectures. *)
Section ArchiDefinitions.

Variable universe_size_bound : nat.

(* TODO : all these type should be defined as inductive tags to avoid spurious*)
(* unfolding of the constants...*)

(* Words are modelled by numbers smaller than universe_size_bound *)
Definition Word := [finType of 'I_universe_size_bound].
(* Adresses are modelled by numbers smaller than universe_size_bound *)
Definition Address := [finType of 'I_universe_size_bound].
(* Values are modelled by numbers smaller than universe_size_bound *)
Definition Value := [finType of 'I_universe_size_bound].
(* Processors are modelled by numbers smaller than universe_size_bound *)
Definition Proc := [finType of 'I_universe_size_bound].
(* Eiid stands for unique event identifier *)
Definition Eiid := [finType of 'I_universe_size_bound].
(* A memory location is specified by an address but we abstract *)
(* this. Can be fixed easily *)
Definition Location := [finType of 'I_universe_size_bound].
(* Program order indexes are modelled by numbers smaller than universe_size_bound. *)
Definition program_order_index := [finType of 'I_universe_size_bound].

(* Iiid stands for instruction identifier : it contains the label of *)
(* the Proc and the program order index identifying uniquely the *)
(* instruction. *)
(*Inductive Iiid : Type := mk_iid : Proc -> program_order_index -> Iiid.*)
Definition Iiid := nosimpl (Proc * program_order_index)%type.

Definition proc_of_Iiid (x : Iiid) := x.1.
Definition poi_of_Iiid (x : Iiid) := x.2.

(* Direction of access : R for reads, W for writes, plus canonical *)
(*structures blurb *)
(* This should as well be defined as a tag of bool *)
Inductive Dirn : Type :=
  | R : Dirn
  | W : Dirn.

Fixpoint Dirn_eqb (x y : Dirn) :=
  match x, y with R, R => true | W, W  => true | _, _ => false end.

Lemma eqDirnP : Equality.axiom Dirn_eqb.
Proof. by do 2 case; constructor. Qed.

Canonical Structure Dirn_eqMixin := EqMixin eqDirnP.
Canonical Structure Dirn_eqType := Eval hnf in EqType Dirn Dirn_eqMixin.


Fixpoint nat_of_Dirn d := match d with R => 0 | _ => 1 end.
Fixpoint nat_oDirn n := match n with 0 => Some R | 1 => Some W | _ =>None end.

Lemma Dirn_pickleK : pcancel nat_of_Dirn nat_oDirn . Proof. by case. Qed.
Definition Dirn_countMixin := CountMixin Dirn_pickleK.
Definition Dirn_choiceMixin := CountChoiceMixin Dirn_countMixin.
Canonical Structure Dirn_choiceType :=
  Eval hnf in ChoiceType Dirn Dirn_choiceMixin.
Canonical Structure Dirn_countType :=
  Eval hnf in CountType Dirn Dirn_countMixin.

Lemma Dirn_enumP : Finite.axiom [:: W; R]. Proof. by case. Qed.
Definition Dirn_finMixin := Eval hnf in FinMixin Dirn_enumP.
Canonical Structure Dirn_finType := Eval hnf in FinType Dirn Dirn_finMixin.


(* An action is an access encoded as a direction (read or write), *)
(* plus a memory location, and a value. *)

Definition Action := nosimpl (Dirn * Location * Value)%type.

Definition dir_of_Action (a : Action) := a.1.1.

Definition loc_of_Action (a : Action) := a.1.2.

Definition value_of_Action (a : Action) := a .2.

Definition Access d l v : Action := (d, l, v).

(* Event:
      - an unique event instance identifier (Eiid); 
      - the associated index in program order and the proc (Iiid); 
      - the associated action (Action) *)


Definition Event := (Eiid * Iiid  * Action)%type.

Definition eiid_of_Event (x : Event) := x.1.1.
Definition iiid_of_Event (x : Event) := x.1.2.
Definition action_of_Event (x : Event) := x.2.


Definition mkEvent e i a : Event := (e, i, a).

Definition dir_of_Event x := dir_of_Action (action_of_Event x).
Definition loc_of_Event x := loc_of_Action (action_of_Event x).
Definition proc_of_Event x := proc_of_Iiid (iiid_of_Event x).
Definition poi_of_Event x := poi_of_Iiid (iiid_of_Event x).

(** Finally, an event structure collects the events of a candidate *)
(**execution. *)

(** ** Event structure: 
       - a (necessarily finite) set of events; 
       - the intra causality relation between events from a same *)
(**instruction. This relation is a total relation on the type Event.*)
Definition Event_struct := 
  nosimpl ({set Event} * (rel Event) * (rel Event))%type.

Definition events_of_ES (x : Event_struct) : {set Event} :=
  let: (evts, _, _) := x in evts.
Definition iico_of_ES (x : Event_struct) : rel Event :=
  let: (_, iico, _) := x in iico.
Definition dp_of_ES (x : Event_struct) : rel Event :=
  let: (_, _, dp) := x in dp.


(* Relates event in es occuring on the same processor *)
Definition same_proc_same_ES (es : Event_struct) : rel Event := 
  [rel e1 e2 \in (events_of_ES es) | proc_of_Event e1 == proc_of_Event e2].

Lemma same_proc_same_ES_eventl : forall es e1 e2, 
  same_proc_same_ES es e1 e2 -> e1 \in events_of_ES es.
Proof. by move=> es e1 e2; do 2 case/andP. Qed.

Lemma same_proc_same_ES_eventr : forall es e1 e2, 
  same_proc_same_ES es e1 e2 -> e2 \in events_of_ES es.

Proof. by move=> es e1 e2; do 2 case/andP. Qed.

(* Program order relation on the event structure es: orders events of *)
(* the event structure es occuring on the same processor accoring to *)
(*their program index *)
Definition po (es : Event_struct) : rel Event := 
  [rel e1 e2 | same_proc_same_ES es e1 e2 && (poi_of_Event e1 <= poi_of_Event e2)].

Lemma po_same_proc_same_ES_sub : forall es, subrel (po es) (same_proc_same_ES es).
Proof.
move=> es x y; rewrite /po /same_proc_same_ES /=.
by case/andP; case/andP=> -> ->.
Qed.

Definition  po_iico_of_ES (E : Event_struct) : rel Event :=  
  relU (po E) (iico_of_ES E).


(*** Predicates for properties of event structures : ***)


Definition same_id_events e :=
  {in (events_of_ES e), 
    injective (fun x => (eiid_of_Event x, iiid_of_Event x))}.

Definition dom_iico_eventsS e :=
  [set x | existsb y : Event, iico_of_ES e x y] \subset (events_of_ES e).

Definition ran_iico_eventsS e :=
  [set x | existsb y : Event, iico_of_ES e y x] \subset (events_of_ES e).

Definition same_line_iico e :=
  subrel [rel x y | poi_of_Event x == poi_of_Event y] (iico_of_ES e).

(* All acyclicity conditions should be localized to events_of_E *)
Definition wf_iico e :=
  [/\ subrel (iico_of_ES e) [rel x y | iiid_of_Event x == iiid_of_Event y],
    (domUrange (iico_of_ES e)) \subset (events_of_ES e) &
    acyclic (iico_of_ES e)
(*    transitive (iico_of_ES e) *)].


Definition reads (es : Event_struct) : pred Event :=
  [pred x \in events_of_ES es | dir_of_Event x == R].

(* do we need  dp_trans : forall E, transitive (dp E) *)
Definition wf_dp E :=
  [/\  subrel (dp_of_ES E) (po_iico_of_ES E) &

    forall x y, dp_of_ES E x y -> reads E x].

Definition wf_ES (e : Event_struct) :=
  [/\same_id_events e,
    wf_dp e &
    wf_iico e].


(* writes of an Event_struct are the events that apear in the Event_struct and whose *)
(* direction are R *)
(* Should this be a set or a pred ? *)
Definition writes (es : Event_struct) : pred Event := 
  [pred x \in events_of_ES es | dir_of_Event x == W].


Definition writes_toS (E : Event_struct) (l : Location) := 
  [set e | 
    [&& e \in (events_of_ES E), (dir_of_Event e == W) & (loc_of_Event e == l)]].

Lemma W_toSP : forall E l e, 
  reflect [/\e \in (events_of_ES E), (dir_of_Event e = W) & (loc_of_Event e = l)] (e \in writes_toS E l).
Proof.
move=> E l e; apply: (iffP idP); rewrite inE.
  by case/and3P=> ->; move/eqP->; move/eqP->.
by case=> -> -> ->; rewrite !eqxx.
Qed.


Definition writes_to_same_loc (s: {set Event}) : rel Event :=
  [rel x y \in s | 
    [&& dir_of_Event x == W, dir_of_Event y == W & loc_of_Event x == loc_of_Event y]].

Lemma W_toS_to_same_loc : forall E l (s : {set Event}) x y, 
  x \in s -> y \in s ->
  x \in writes_toS E l -> y \in writes_toS E l -> writes_to_same_loc s x y.
Proof.
move=> E l s x y sx sy; case/W_toSP=> inEx Wx lx; case/W_toSP=> inEy Wy ly. 
by rewrite /writes_to_same_loc /= sx sy Wx Wy lx ly !eqxx.
Qed.

Lemma W_to_same_locP : forall (s : {set Event})(x y : Event),
  reflect [/\ x \in s, y \in s, 
    dir_of_Event x = W, dir_of_Event y = W & loc_of_Event x = loc_of_Event y]
    (writes_to_same_loc s x y).
Proof.
move=> s x y; apply: (iffP idP); rewrite /writes_to_same_loc /=.
  by case/andP; case/andP=> -> ->; case/and3P; move/eqP->; move/eqP->; move/eqP.
by case=> -> -> -> -> ->; rewrite !eqxx.
Qed.

Lemma W_to_same_loc_trans : forall s, transitive (writes_to_same_loc s).
Proof.
move=> s x y z; rewrite {3}/writes_to_same_loc /=. 
case/W_to_same_locP=> -> _ -> _ ->; case/W_to_same_locP=> _ -> _ -> <-.
by rewrite !eqxx.
Qed.

Lemma W_to_same_loc_sym : forall s, symmetric (writes_to_same_loc s).
Proof.
move=> s x y. 
rewrite /writes_to_same_loc /=; congr (_ && _); first by rewrite andbC.
by rewrite !andbA; congr (_ && _); rewrite 1?andbC // eq_sym.
Qed.


(* More inclusion lemmas here ? *)


Definition Execution_witness := nosimpl (rel Event * rel Event)%type.

Definition mkew ws rf : Execution_witness := (ws, rf).
Definition ws_of_EW (e : Execution_witness) := e.1.
Definition rf_of_EW (e : Execution_witness) := e.2.


(* Stores in an Event_struct *)

(* reads of an Event_struct are the events that apear in the Event_struct and whose *)
(* direction are R *)

Definition wsmaps (e1 e2 : Event) :=
  [&& dir_of_Event e1 == W, dir_of_Event e2 == W & loc_of_Event e1 == loc_of_Event e2].

Lemma wsmaps_isWl :  forall e1 e2, wsmaps e1 e2 -> dir_of_Event e1 = W.
Proof. by move=> e1 e2; case/andP; move/eqP. Qed.

Lemma wsmaps_isWr :  forall e1 e2, wsmaps e1 e2 -> dir_of_Event e2 = W.
Proof. by move=> e1 e2; case/and3P=> _; move/eqP. Qed.

Lemma wsmaps_loc : forall e1 e2, wsmaps e1 e2 ->
  loc_of_Event e1 = loc_of_Event e2.
Proof. by move=> e1 e2; case/and3P=> _ _; move/eqP. Qed.

Definition wf_ws (E : Event_struct)(s : {set Event})(ws : rel Event) :=
 [/\
    forall l : Location, transitive [rel x y \in  (writes_toS E l) | ws x y],
    forall l : Location, irreflexive [rel x y \in  (writes_toS E l) | ws x y]&
(* update   forall l : Location, linear_on ws  (writes_toS E l)] /\*)
      forall l : Location, {in (writes_toS E l) &, total_irr ws}] /\
 [/\   
   forall e1 e2, ws e1 e2 -> e1 \in s,
   forall e1 e2, ws e1 e2 -> e2 \in s &
   forall e1 e2, ws e1 e2 -> wsmaps e1 e2].

Definition read_from (e : Event) (l : Location) :=
  (dir_of_Event e == R) && (loc_of_Event e == l).

Definition rfmaps (e1 e2 : Event) :=
  [&& dir_of_Event e1 == W, dir_of_Event e2 == R & loc_of_Event e1 == loc_of_Event e2].

Lemma rfmaps_isWl :  forall e1 e2, rfmaps e1 e2 -> dir_of_Event e1 = W.
Proof. by move=> e1 e2; case/andP; move/eqP. Qed.

Lemma rfmaps_isRr :  forall e1 e2, rfmaps e1 e2 -> dir_of_Event e2 = R.
Proof. by move=> e1 e2; case/and3P=> _; move/eqP. Qed.

Lemma rfmaps_loc : forall e1 e2, rfmaps e1 e2 ->
  loc_of_Event e1 = loc_of_Event e2.
Proof. by move=> e1 e2; case/and3P=> _ _; move/eqP. Qed.

Lemma rfmaps_irrefl : irreflexive rfmaps.
Proof.
move=> x; case abs : (rfmaps x x) => //.
by move: (rfmaps_isWl abs); rewrite (rfmaps_isRr abs).
Qed.


(* rf is in fact a relation on Events modeling Rfmap *)
Definition wf_rf  (s: {set Event}) (rf : rel Event) :=
  [/\ (forall er : Event , er \in s -> exists ew : Event, rf ew er), (*Hrf_init*) forall x w1 w2, rf w1 x -> rf w2 x -> w1 = w2,  (*Hrf_uni*)
  forall e1 e2, rf e1 e2 -> e1 \in s,
  forall e1 e2, rf e1 e2 -> e2 \in s &
  forall e1 e2, rf e1 e2 -> (rfmaps e1 e2) (*Hrf_cands*) 
   ]. 

Definition wf_EW (E : Event_struct)(X : Execution_witness) :=
  (wf_ws E (events_of_ES E) (ws_of_EW X)) /\
  (wf_rf (events_of_ES E) (rf_of_EW X)).


Definition fr (X : Execution_witness) : rel Event :=
  [rel er ew : Event | 
    existsb wr : Event, rf_of_EW X wr er && ws_of_EW X wr ew].


Record Archi : Type := mkArchi{
  ppo : Event_struct -> rel Event;
  grf :   Execution_witness -> rel Event;
  abc   : Event_struct -> Execution_witness -> rel Event;
  grf_valid : forall X, subrel (grf X) (rf_of_EW X);
  ppo_valid : forall E, subrel (ppo E) (po_iico_of_ES E);
  ab_evts : forall E X, 
    subrel (abc E X) [rel x y \in (events_of_ES E)]
}.

(* Processor issue order : this is a minimal requirement, cf. alpha *)
(*model*)
Definition po_loc (e : Event_struct) : rel Event :=
  [rel x y | (loc_of_Event x == loc_of_Event y) && (po_iico_of_ES e x y)].



(** * AB orderings *)

(* abc is an AB ordering *)

Definition ab_well_formed (s : pred Event) (abc : rel Event) :=
  subrel abc [rel x y \in s | x != y].


(** ** Initial and final states *)
Definition init (X : Execution_witness) : pred Event :=
  [pred x | 
    existsb wx : Event, (rf_of_EW X) wx x && ~~(existsb ew : Event, (ws_of_EW X) ew wx)].

Definition final (X : Execution_witness) : pred Event :=
  [pred x | ~~(existsb e : Event, (ws_of_EW X) x e)].

(** * Finally, definition of a valid execution *)
  
(** ** happens before *)

Definition com  (X : Execution_witness) : rel Event :=
  relU (relU (rf_of_EW X) (fr X)) (ws_of_EW X).

Definition sc_check (E : Event_struct) (X : Execution_witness) : Prop :=
  acyclic (relU (po_iico_of_ES E) (com X)).

Definition compete E : rel Event :=
  [rel e1 e2 \in (events_of_ES E)|
  [&& (loc_of_Event e1 == loc_of_Event e2), 
      (proc_of_Event e1 != proc_of_Event e2) &
      (writes E e1 || writes E e2)]].

(* Architecture dependant definitions *)
Section Wmm.

Variable A : Archi.

Definition ghb  (E : Event_struct) (X : Execution_witness) : 
  rel Event :=
  relU (grf A X) ((relU (abc A E X)
      (relU (relU (ws_of_EW X) (fr X)) (ppo A E)))).

Definition valid_execution (E : Event_struct) (X : Execution_witness) : Prop :=
  [/\
  acyclic (relU (com X) (po_loc E)), (* uniproc: per location coherence *) 
  acyclic (relU (rf_of_EW X) (dp_of_ES E)) & (* thin *)
  acyclic (ghb E X)].

End Wmm.

End ArchiDefinitions.

